import { Injectable } from '@nestjs/common';
import { PrismaClientKnownRequestError } from '@prisma/client/runtime';
import { Point } from '../../domain/point.type';
import { DatabaseException } from '../../exceptions/database.exception';
import { ICollection } from '../../interfaces/collection.interface';
import { IRepository } from '../../interfaces/repository.interface';
import { PrismaService } from './prisma-service';
import { Territory } from 'src/modules/territories/domain/entities/territory';

/**
 * Child classes MUST redefined _model property with appropriate model name
 */
@Injectable()
export abstract class PrismaRepository<T> implements IRepository<T> {
  protected _model: string;

  constructor(protected readonly _prisma: PrismaService) {}

  async findAll(
    page = 1,
    perPage = 10,
    where?: any,
    include?: any,
  ): Promise<ICollection<T>> {
    const [data, total] = await this._prisma.$transaction([
      this._prisma[this._model].findMany({
        where,
        include,
        skip: (page - 1) * perPage,
        take: perPage,
      }),
      this._prisma[this._model].count({
        where,
      }),
    ]);
    return Promise.resolve({
      data,
      total,
    });
  }

  async findOneByUuid(uuid: string): Promise<T> {
    try {
      const entity = await this._prisma[this._model].findUnique({
        where: { uuid },
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async findOne(where: any, include?: any): Promise<T> {
    try {
      const entity = await this._prisma[this._model].findFirst({
        where: where,
        include: include,
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(PrismaClientKnownRequestError.name, e.code);
      } else {
        throw new DatabaseException();
      }
    }
  }

  // TODO : using any is not good, but needed for nested entities
  // TODO : Refactor for good clean architecture ?
  async create(entity: Partial<T> | any, include?: any): Promise<T> {
    try {
      const res = await this._prisma[this._model].create({
        data: entity,
        include: include,
      });

      return res;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async update(uuid: string, entity: Partial<T>): Promise<T> {
    try {
      const updatedEntity = await this._prisma[this._model].update({
        where: { uuid },
        data: entity,
      });
      return updatedEntity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async updateWhere(
    where: any,
    entity: Partial<T> | any,
    include?: any,
  ): Promise<T> {
    try {
      const updatedEntity = await this._prisma[this._model].update({
        where: where,
        data: entity,
        include: include,
      });

      return updatedEntity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async delete(uuid: string): Promise<T> {
    try {
      const entity = await this._prisma[this._model].delete({
        where: { uuid },
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async deleteMany(where: any): Promise<void> {
    try {
      const entity = await this._prisma[this._model].deleteMany({
        where: where,
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async findForPoint(point: Point): Promise<ICollection<T>> {
    const strPoint = `SELECT uuid, name FROM ${this._model} WHERE ST_Intersects(ST_GeomFromText('POINT(${point.lon} ${point.lat})',4326),shape) = true`;
    const territories: Array<T> = await this._prisma.$queryRawUnsafe(strPoint);
    return Promise.resolve({
      data: territories,
      total: territories.length,
    });
  }

  async createTerritory(territory: Territory): Promise<T> {
    const command = `INSERT INTO ${this._model} VALUES ('bb281075-1b98-4456-89d6-c643d3044a91','${territory.name}', ST_GeomFromGeoJSON('{"type":"MultiPolygon","coordinates":${territory.shape}}'),'2023-02-07 15:58:00','2023-02-07 15:58:00')`;
    const affectedRowNumber = await this._prisma.$executeRawUnsafe(command);
    if (affectedRowNumber == 1) {
      return this.findOne({
        name: territory.name,
      });
    }
    throw new DatabaseException();
  }
}
